<?php

    namespace wpsg;

    /**
     * Neue Klasse zur Berechnung des Warenkorbes
     * Soll die alte basket Klasse ersetzen
     * 
     */
    class wpsg_calculation {

        /**
         * @var wpsg_db
         */
        private $db = null;

        /**
         * @var \wpsg_ShopController
         */
        private $shop = null;

        /**
         * @var null 
         */
        private $arCalculation = null;
        
        /** @var array */
        private $arProducts = [];
        
        /** @var array */
        private $shipping = null;
        
        /** @var array */
        private $payment = null;
        
        /** @var \wpsg_country */
        private $oCountry = null;
        
        /** @var array */
        private $arDiscount = null;
        
        public function __construct() {
            
            $this->db = &$GLOBALS['wpsg_db'];
            $this->shop = &$GLOBALS['wpsg_sc'];
            
            $this->oCountry = $this->shop->getDefaultCountry();
            
        }        
        
        /** 
         * Setzt das Standardland für die Steuerberechnung wenn im $tax_key kein Land definiert ist 
         */
        public function setCountry($oCountry) {
            
            $this->oCountry = $oCountry;
            
        }
        
        public function removeDiscount() {
            
            $this->arDiscount = null;
            
        }
        
        public function addDiscount($discount_value, $discount_tax_key, $discount_set) {
            
            $this->arDiscount = [
                [
                    'set' => $discount_set,
                    'value' => $discount_value,
                    'tax_key' => $this->normalizeTaxKey($discount_tax_key)
                ]
            ];
                        
        }
        
        public function addProduct($product_key, $product_index, $price_netto, $amount, $tax_key, $order_product_id = false) {
            
            $tax_key = $this->normalizeTaxKey($tax_key);
          
            $p_set = [
                'product_key' => $product_key,
                'product_id' => $this->shop->getProduktId($product_key),
                'product_index' => $product_index,
                'netto' => $price_netto,
                'amount' => $amount,
                'tax_key' => $tax_key,
                'order_product_id' => $order_product_id
            ];
            
            if (wpsg_isSizedInt($order_product_id)) {
                
                // Ersetzen wenn vorhanden
                foreach ($this->arProducts as $k => $p) {
                    
                    if ($p['order_product_id'] == $order_product_id) {
                        
                        $this->arProducts[$k] = $p_set;
                        return;
                        
                    }
                    
                }
                
            }
            
            $this->arProducts[] = $p_set;
            
        }
        
        public function removeProduct($order_product_id) {
            
            foreach ($this->arProducts as $k => $p) {
                
                if ($p['order_product_id'] == $order_product_id) unset($this->arProducts[$k]);
                
            }
            
        }
        
        /**
         * Setzt die Versandkosten
         */
        public function setShipping($key, $netto, $tax_key) {

            $tax_key = $this->normalizeTaxKey($tax_key);
            
            $this->shipping = [
                'key' => $key,
                'netto' => $netto,
                'brutto' => 0,
                'tax_key' => $tax_key
            ];
                         
        }

        /**
         * Setzt die Zahlungskosten
         */
        public function setPayment($key, $netto, $tax_key) {

            $tax_key = $this->normalizeTaxKey($tax_key);
            
            $this->payment = [
                'key' => $key,
                'netto' => $netto,
                'brutto' => 0,
                'tax_key' => $tax_key
            ];
            
        }

        public function getCalculationArray($taxView = false, $force_rebuild = false) {

            if ($this->oCountry === null) throw new \Exception(__('Warenkorb kann nicht ohne ein gesetztes Land berechnet werden.', 'wpsg'));
            if ($this->shipping === null) throw new \Exception(__('Kein Versand definiert.'));
            if ($this->payment === null) throw new \Exception(__('Keine Zahlung definiert.'));

            if ($force_rebuild === true || is_null($this->arCalculation)) {

                $this->arCalculation = [
                    'land' => $this->db->fetchRow("SELECT * FROM `".WPSG_TBL_LAND."` WHERE `id` = '".wpsg_q($this->oCountry->id)."' "),
                    'products' => $this->arProducts
                ];

                $this->arCalculation['discount'] = [];                
                $this->arCalculation['sum'] = [
                    'product_netto' => 0,
                    'product_brutto' => 0,
                    'prediscount_netto' => 0,
                    'prediscount_brutto' => 0,
                    'netto' => 0,
                    'brutto' => 0,
                    'tax' => 0,
                    'discount_netto' => 0,
                    'discount_brutto' => 0
                ];

                $this->arCalculation['products'] = $this->arProducts;
                
                foreach ($this->arProducts as $k => $p) {
                    
                    $amount_price_netto = $p['netto'] * $p['amount'];
                    
                    $this->arCalculation['sum']['product_netto'] += $amount_price_netto;
                    $this->arCalculation['sum']['netto'] += $amount_price_netto;

                    $this->addTax($amount_price_netto, $p['tax_key']);

                }

                $this->arCalculation['shipping'] = $this->shipping;
                $this->arCalculation['payment'] = $this->payment;

                $this->arCalculation['sum']['netto'] += $this->shipping['netto'];
                $this->addTax($this->shipping['netto'], $this->shipping['tax_key']);

                $this->arCalculation['sum']['netto'] += $this->payment['netto'];
                $this->addTax($this->payment['netto'], $this->payment['tax_key']);

                // Rabatte einbeziehen
                $this->calculateDiscount();
                
                // Jetzt den Bruttopreis berechnen
                $this->calculateTax();

            }

            if ($taxView !== false) {
            
                $this->setTaxView($taxView);
            
            }
                
            return $this->arCalculation;

        }

        /**
         * Speichert die Berechnung in eine Bestellung
         * 
         * @param bool $id
         */
        public function toDB($id = false, $db_data = []) {
           
            $arCalculation = $this->getCalculationArray();

            $db_data['price_gesamt_netto'] = wpsg_q($arCalculation['sum']['netto']);
            $db_data['price_shipping_netto'] = wpsg_q($this->arCalculation['shipping']['netto']);
            $db_data['price_payment_netto'] = wpsg_q($this->arCalculation['payment']['netto']);
            
            $db_data['price_gesamt_brutto'] = wpsg_q($arCalculation['sum']['brutto']);
            $db_data['price_shipping_brutto'] = wpsg_q($this->arCalculation['shipping']['brutto']);
            $db_data['price_payment_brutto'] = wpsg_q($this->arCalculation['payment']['brutto']);
            
            if ($this->shop->getFrontendTaxview() === WPSG_NETTO) {
                
                $db_data['price_gesamt'] = wpsg_q($arCalculation['sum']['netto']);
                $db_data['price_shipping'] = wpsg_q($arCalculation['shipping']['netto']);
                $db_data['price_payment'] = wpsg_q($arCalculation['payment']['netto']);
                $db_data['price_rabatt'] = wpsg_q($arCalculation['sum']['discount_netto']);
                
            } else {

                $db_data['price_gesamt'] = wpsg_q($arCalculation['sum']['brutto']);
                $db_data['price_shipping'] = wpsg_q($arCalculation['shipping']['brutto']);
                $db_data['price_payment'] = wpsg_q($arCalculation['payment']['brutto']);
                $db_data['price_rabatt'] = wpsg_q($arCalculation['sum']['discount_brutto']);
                
            }
            
            
            $db_data['price_rabatt_netto'] = wpsg_q($arCalculation['sum']['discount_netto']);
            $db_data['price_rabatt_brutto'] = wpsg_q($arCalculation['sum']['discount_brutto']);
            
            if (wpsg_isSizedArray($this->arDiscount)) {
            
                foreach ($this->arDiscount as $d) {
                    
                    $db_data['discount_tax_key'] = $d['tax_key'];
                    $db_data['discount_set'] = $d['set'];
                    
                }
                    
            }
                        
            if (wpsg_isSizedInt($id)) {
                
                $this->db->UpdateQuery(WPSG_TBL_ORDER, $db_data, " `id` = '".wpsg_q($id)."' ");
                
            }  else {
                
                $id = $this->db->ImportQuery(WPSG_TBL_ORDER, $db_data);
                
                // TODO: Land zuweisen und speichern
                
            }

            $arOrderProductID = [];
            
            // Produkte speichern
            foreach ($arCalculation['products'] as $p) {
                
                $db_data = [
                    'o_id' => wpsg_q($id),
                    'p_id' => wpsg_q($p['product_id']),
                    'productkey' => wpsg_q($p['product_key']),
                    'product_index' => wpsg_q($p['product_index']),
                    'menge' => wpsg_q($p['amount']),
                    'price_netto' => wpsg_q($p['netto']),
                    'price_brutto' => wpsg_q($p['brutto']),
                    'mwst_key' => wpsg_q($p['tax_key'])
                ];
                
                if (wpsg_isSizedInt($p['order_product_id'])) {
                    
                    $arOrderProductID[] = $p['order_product_id'];
                    
                    $this->db->UpdateQuery(WPSG_TBL_ORDERPRODUCT, $db_data, " `id` = '".wpsg_q($p['order_product_id'])."' ");
                    
                } else {

                    $arOrderProductID[] = $this->db->ImportQuery(WPSG_TBL_ORDERPRODUCT, $db_data);
                    
                }
                
            }
            
            if (wpsg_isSizedArray($arOrderProductID)) {
                
                $this->db->Query("DELETE FROM `".WPSG_TBL_ORDERPRODUCT."` WHERE `o_id` = '".wpsg_q($id)."' AND `id` NOT IN (".wpsg_q(implode(',', $arOrderProductID)).") ");
                
            }
                
            return $id;
            
        } // public function toDB($id = false)

        public function fromDB($id) {
            
            $db_order = $this->db->fetchRow("SELECT O.*, A.`land` FROM `".WPSG_TBL_ORDER."` AS O LEFT JOIN `".WPSG_TBL_ADRESS."` AS A ON (O.`adress_id` = A.`id`) WHERE O.`id` = '".wpsg_q($id)."' ");
            $db_products = $this->db->fetchAssoc("SELECT * FROM `".WPSG_TBL_ORDERPRODUCT."` WHERE `o_id` = '".wpsg_q($id)."' ");
            
            $this->setCountry(\wpsg_country::getInstance($db_order['land']));
            $this->setShipping($db_order['type_shipping'], $db_order['price_shipping_netto'],$db_order['shipping_tax_key']);
            $this->setPayment($db_order['type_payment'], $db_order['price_payment_netto'], $db_order['payment_tax_key']);
            
            foreach ($db_products as $db_p) {
               
                $this->addProduct($db_p['productkey'], $db_p['product_index'], $db_p['price_netto'], $db_p['menge'],$db_p['mwst_key'], $db_p['id']);
                
            }
            
            // Rabatte
            if (wpsg_tf($db_order['price_rabatt_netto']) > 0) {
                
                $this->addDiscount($db_order['price_rabatt_netto'], $db_order['discount_tax_key'], $db_order['discount_set']);
                
            }
            
            
        } // public function fromDB($id)

        /**
         * Gibt einen Array zurück, in dem die beteiligten Steuersätze mit Länderkürzel definiert sind
         */
        public function getTaxLabelArray($short = false) {

            $arTaxLabel = [];
            $arTaxGgroup = wpsg_tax_groups(false, true);            
            $arCalculation = $this->getCalculationArray();
            
            foreach ($arCalculation['tax'] as $tax) {
                                    
                if (preg_match('/_\d+/', $tax['key'])) {
                    
                    $arTaxKey = explode('_', $tax['key']);
                    $tax_key = $arTaxKey[0];
                    
                    $oCountry = \wpsg_country::getInstance($arTaxKey[1]);
                    
                } else {
                    
                    $tax_key = $tax['key'];
                    $oCountry = $this->shop->getDefaultCountry();
                                        
                }

                $tax_value = $oCountry->getTax($tax_key);

                if ($short === true) {

                    if ($tax_key !== '0') $arTaxLabel[$tax['key']] = wpsg_ff(wpsg_tf($tax_value), '%');
                    else $arTaxLabel[$tax['key']] = 'anteilig';

                } else {
                
                    if ($tax_key !== '0') $arTaxLabel[$tax['key']] = $arTaxGgroup[$tax_key].' ('.wpsg_ff(wpsg_tf($tax_value), '%').' / '.$oCountry->kuerzel.')';
                    else $arTaxLabel[$tax['key']] = $arTaxGgroup[$tax_key];
                    
                }
                                    
            }
            
            return $arTaxLabel;
                
        } 
        
        public function getMaxProductIndex() {
            
            $max_index = 0;
            
            foreach ($this->arProducts as $p) {
                
                $max_index = max($max_index, $p['product_index']);
                
            }
            
            return max($max_index, sizeof($this->arProducts) - 1);
            
        }
        
        /* Private */

        private function setTaxView($taxView) {

            $this->getCalculationArray();

            if ($taxView === WPSG_BRUTTO) {

                $suffix = 'brutto';

            } else {

                $suffix = 'netto';

            }

            $this->arCalculation['sum']['product'] = $this->arCalculation['sum']['product_'.$suffix];
            $this->arCalculation['sum']['display'] = $this->arCalculation['sum'][$suffix];
            $this->arCalculation['sum']['shipping'] = $this->arCalculation['shipping'][$suffix];
            $this->arCalculation['sum']['payment'] = $this->arCalculation['payment'][$suffix];

            $this->arCalculation['shipping']['price'] = $this->arCalculation['shipping'][$suffix];
            $this->arCalculation['payment']['price'] = $this->arCalculation['payment'][$suffix];
            $this->arCalculation['sum']['discount_display'] = $this->arCalculation['sum']['discount_'.$suffix];
            $this->arCalculation['sum']['prediscount_display'] = $this->arCalculation['sum']['prediscount_'.$suffix];            
            
            foreach ($this->arCalculation['products'] as $k => $p) {

                $this->arCalculation['products'][$k]['price'] = $this->arCalculation['products'][$k][$suffix];
                $this->arCalculation['products'][$k]['sum_display'] = $this->arCalculation['products'][$k]['sum_'.$suffix];

            }

        }

        /**
         * Bezieht die Rabatte in die Berechnung ein
         */
        private function calculateDiscount() {

            $this->arCalculation['sum']['prediscount_netto'] = $this->arCalculation['sum']['netto'];
            $this->arCalculation['sum']['discount_netto'] = 0;
            
            if (wpsg_isSizedArray($this->arDiscount)) {

                $this->arCalculation['discount'] = $this->arDiscount;
                                                
                foreach ($this->arDiscount as $k => $d) {
                    
                    if (strpos($d['value'], '%') !== false) {
                        
                        // Prozentualer Wert
                        $discount_netto = $this->arCalculation['sum']['netto'] / 100 * wpsg_tf($d['value']);
                        
                    } else {
                        
                        // Absoluter Netto Wert
                        $discount_netto = wpsg_tf($d['value']);
                        
                    }
                    
                    $this->arCalculation['discount'][$k]['netto'] = $discount_netto;                    
                    $this->arCalculation['tax'][$d['tax_key']]['netto'] -= $discount_netto;
                    
                    $this->arCalculation['sum']['discount_netto'] += $discount_netto;
                    $this->arCalculation['sum']['netto'] -= $discount_netto;
                    
                }
                                
            }
            
        } // private function calculateDiscount()
        
        /** 
         * Berechnet den Steueranteil 
         */
        private function calculateTax() {
            
            if ($this->arCalculation === null) throw new \Exception(__('Warenkorb nicht gefüllt.', 'wpsg'));
            
            // Den Anteiligen Satz ermitteln
            if (array_key_exists('0', $this->arCalculation['tax'])) {
                
                $sum_netto = $this->arCalculation['sum']['netto'];
                
                // Den Anteiligen Netto Betrag beachte ich nicht bei der Anteilermittlung
                $sum_netto -= $this->arCalculation['tax'][0]['netto'];
                
                foreach ($this->arCalculation['tax'] as $tax_key => $tax) {
                    
                    if ($tax_key == '0') $this->arCalculation['tax'][$tax_key]['part'] = 0;
                    else $this->arCalculation['tax'][$tax_key]['part'] = $tax['netto'] / $sum_netto;
                    
                }
                                
            }
                 
            foreach ($this->arCalculation['products'] as $k => $p) {

                $this->arCalculation['products'][$k]['brutto'] = $this->calculateTaxPart($p['netto'], $p['tax_key']);
                
                if ($this->arCalculation['products'][$k]['brutto'] > 0) $this->arCalculation['products'][$k]['sum_tax'] = $this->arCalculation['products'][$k]['brutto'] - $this->arCalculation['products'][$k]['netto'];
                else $this->arCalculation['products'][$k]['sum_tax'] = 0;

                $this->arCalculation['tax'][$p['tax_key']]['sum_tax'] += $this->arCalculation['products'][$k]['sum_tax'] * $this->arCalculation['products'][$k]['amount'];                 
                $this->arCalculation['sum']['tax'] += $this->arCalculation['products'][$k]['sum_tax'] * $this->arCalculation['products'][$k]['amount'];
                
                $this->arCalculation['sum']['brutto'] += $this->arCalculation['products'][$k]['brutto'] * $this->arCalculation['products'][$k]['amount'];
                $this->arCalculation['sum']['prediscount_brutto'] += $this->arCalculation['products'][$k]['brutto'] * $this->arCalculation['products'][$k]['amount'];
                $this->arCalculation['sum']['product_brutto'] += $this->arCalculation['products'][$k]['brutto'] * $this->arCalculation['products'][$k]['amount'];
                
            }

            // Versand
            $this->arCalculation['shipping']['brutto'] = $this->calculateTaxPart($this->arCalculation['shipping']['netto'], $this->arCalculation['shipping']['tax_key']);
            $this->arCalculation['sum']['brutto'] += $this->arCalculation['shipping']['brutto'];
            $this->arCalculation['sum']['prediscount_brutto'] += $this->arCalculation['shipping']['brutto'];

            if ($this->arCalculation['shipping']['brutto'] > 0) $this->arCalculation['shipping']['sum_tax'] = $this->arCalculation['shipping']['brutto'] - $this->arCalculation['shipping']['netto'];
            else $this->arCalculation['shipping']['sum_tax'] = 0;

            $this->arCalculation['tax'][$this->shipping['tax_key']]['brutto'] += $this->arCalculation['shipping']['brutto'];
            $this->arCalculation['tax'][$this->shipping['tax_key']]['sum_tax'] += $this->arCalculation['shipping']['sum_tax'];
            
            // Zahlung
            $this->arCalculation['payment']['brutto'] = $this->calculateTaxPart($this->arCalculation['payment']['netto'], $this->arCalculation['payment']['tax_key']);
            $this->arCalculation['sum']['brutto'] += $this->arCalculation['payment']['brutto'];
            $this->arCalculation['sum']['prediscount_brutto'] += $this->arCalculation['payment']['brutto'];

            if ($this->arCalculation['payment']['brutto'] > 0) $this->arCalculation['payment']['sum_tax'] = $this->arCalculation['payment']['brutto'] - $this->arCalculation['payment']['netto'];
            else $this->arCalculation['payment']['sum_tax'] = 0;

            $this->arCalculation['tax'][$this->payment['tax_key']]['brutto'] += $this->arCalculation['payment']['brutto'];
            $this->arCalculation['tax'][$this->payment['tax_key']]['sum_tax'] += $this->arCalculation['payment']['sum_tax'];
            
            // Rabatte
            foreach ($this->arCalculation['discount'] as $k => $d) {
                
                $this->arCalculation['discount'][$k]['brutto'] = $this->calculateTaxPart($d['netto'], $d['tax_key']);

                if ($this->arCalculation['discount'][$k]['brutto'] > 0) $this->arCalculation['discount'][$k]['sum_tax'] = $this->arCalculation['discount'][$k]['brutto'] - $this->arCalculation['discount'][$k]['netto'];
                else $this->arCalculation['discount'][$k]['sum_tax'] = 0;

                $this->arCalculation['tax'][$d['tax_key']]['sum_tax'] -= $this->arCalculation['discount'][$k]['sum_tax'];

                $this->arCalculation['sum']['tax'] -= $this->arCalculation['discount'][$k]['sum_tax'];
                $this->arCalculation['sum']['brutto'] -= $this->arCalculation['discount'][$k]['brutto'];                
                $this->arCalculation['sum']['discount_brutto'] += $this->arCalculation['discount'][$k]['brutto'];
                
            }
            
            // Gesamt
            if ($this->arCalculation['sum']['brutto'] > 0) $this->arCalculation['sum']['tax'] = $this->arCalculation['sum']['brutto'] - $this->arCalculation['sum']['netto'];
            else $this->arCalculation['sum']['tax'] = 0;
            
            // Anteilig noch auf die Steuersätze verteilen
            if (isset($this->arCalculation['tax'][0]['sum_tax'])) {
                
                $part_sum_tax = $this->arCalculation['tax'][0]['sum_tax'];
                $part_netto = $this->arCalculation['tax'][0]['netto'];
                $part_brutto = $this->arCalculation['tax'][0]['brutto'];
                
                foreach ($this->arCalculation['tax'] as $tax_key => $tax_data) {
                    
                    $this->arCalculation['tax'][$tax_key]['sum_tax'] += $part_sum_tax * $tax_data['part'];
                    $this->arCalculation['tax'][$tax_key]['netto'] += $part_netto * $tax_data['part'];
                    $this->arCalculation['tax'][$tax_key]['brutto'] += $part_brutto * $tax_data['part'];
                    
                }
                
            }
            
        }

        /**
         * Berechnet den Brutto Wert zu einem Nettowert
         */
        private function calculateTaxPart($netto, $tax_key) {
            
            if ($tax_key == '0') {
                
                $brutto = 0;
                
                foreach ($this->arCalculation['tax'] as $tax_key => $tax) {
                    
                    if ($tax_key != '0') {
                        
                        $brutto += $netto * $tax['part'] * (1 + $tax['tax_value'] / 100);
                        
                    }
                    
                }
                
                return $brutto;
                
            } else {

                return $netto * (1 + $this->arCalculation['tax'][$tax_key]['tax_value'] / 100);
                
            }
            
        }

        /**
         * Im Array sollen die Tax Schlüssel immer mit Land gespeichert sein
         */
        private function normalizeTaxKey($tax_key) {

            if ($tax_key == '') $tax_key = '0';
            
            if ($tax_key != '0' && !preg_match('/\_\d+$/', $tax_key)) return $tax_key.'_'.$this->oCountry->id;
            else return $tax_key;
            
        }
        
        private function addTax($price_netto, $tax_key) {
            
            if (!isset($this->arCalculation['tax'][$tax_key])) {
                   
                $arTaxKey = explode('_', $tax_key);
                
                $country = \wpsg_country::getInstance($arTaxKey[1]);
                $tax_value = $country->getTax($arTaxKey[0]);
                
                $this->arCalculation['tax'][$tax_key] = [
                    'key' => $tax_key,
                    'netto' => $price_netto,
                    'brutto' => 0,
                    'tax_value' => $tax_value,
                    'sum_tax' => 0
                ];
                
            } else {
                
                $this->arCalculation['tax'][$tax_key]['netto'] += $price_netto;
                
            }
            
        }

    }
         
         